# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from unittest.mock import AsyncMock, Mock, patch

from aioresponses import aioresponses  # type: ignore
import pytest  # type: ignore
import pytest_asyncio  # type: ignore

from google.auth import exceptions
from google.auth.aio import _helpers as _helpers_async
import google.auth.aio.transport.aiohttp as auth_aiohttp

try:
    import aiohttp  # type: ignore
except ImportError as caught_exc:  # pragma: NO COVER
    raise ImportError(
        "The aiohttp library is not installed from please install the aiohttp package to use the aiohttp transport."
    ) from caught_exc


@pytest.fixture
def mock_response():
    response = Mock()
    response.status = 200
    response.headers = {"Content-Type": "application/json", "Content-Length": "100"}
    mock_iterator = AsyncMock()
    mock_iterator.__aiter__.return_value = iter(
        [b"Cavefish ", b"have ", b"no ", b"sight."]
    )
    response.content.iter_chunked = lambda chunk_size: mock_iterator
    response.read = AsyncMock(return_value=b"Cavefish have no sight.")
    response.close = AsyncMock()

    return auth_aiohttp.Response(response)


class TestResponse(object):
    @pytest.mark.asyncio
    async def test_response_status_code(self, mock_response):
        assert mock_response.status_code == 200

    @pytest.mark.asyncio
    async def test_response_headers(self, mock_response):
        assert mock_response.headers["Content-Type"] == "application/json"
        assert mock_response.headers["Content-Length"] == "100"

    @pytest.mark.asyncio
    async def test_response_content(self, mock_response):
        content = b"".join([chunk async for chunk in mock_response.content()])
        assert content == b"Cavefish have no sight."

    @pytest.mark.asyncio
    async def test_response_content_raises_error(self, mock_response):
        with patch.object(
            mock_response._response.content,
            "iter_chunked",
            side_effect=aiohttp.ClientPayloadError,
        ):
            with pytest.raises(exceptions.ResponseError) as exc:
                [chunk async for chunk in mock_response.content()]
            exc.match("Failed to read from the payload stream")

    @pytest.mark.asyncio
    async def test_response_read(self, mock_response):
        content = await mock_response.read()
        assert content == b"Cavefish have no sight."

    @pytest.mark.asyncio
    async def test_response_read_raises_error(self, mock_response):
        with patch.object(
            mock_response._response,
            "read",
            side_effect=aiohttp.ClientResponseError(None, None),
        ):
            with pytest.raises(exceptions.ResponseError) as exc:
                await mock_response.read()
            exc.match("Failed to read the response body.")

    @pytest.mark.asyncio
    async def test_response_close(self, mock_response):
        await mock_response.close()
        mock_response._response.close.assert_called_once()

    @pytest.mark.asyncio
    async def test_response_content_stream(self, mock_response):
        itr = mock_response.content().__aiter__()
        content = []
        try:
            while True:
                chunk = await itr.__anext__()
                content.append(chunk)
        except StopAsyncIteration:
            pass
        assert b"".join(content) == b"Cavefish have no sight."


@pytest.mark.asyncio
class TestRequest:
    @pytest_asyncio.fixture
    async def aiohttp_request(self):
        request = auth_aiohttp.Request()
        yield request
        await request.close()

    async def test_request_call_success(self, aiohttp_request):
        with aioresponses() as m:
            mocked_chunks = [b"Cavefish ", b"have ", b"no ", b"sight."]
            mocked_response = b"".join(mocked_chunks)
            m.get("http://example.com", status=200, body=mocked_response)
            response = await aiohttp_request("http://example.com")
            assert response.status_code == 200
            assert response.headers == {"Content-Type": "application/json"}
            content = b"".join([chunk async for chunk in response.content()])
            assert content == b"Cavefish have no sight."

    async def test_request_call_success_with_provided_session(self):
        mock_session = aiohttp.ClientSession()
        request = auth_aiohttp.Request(mock_session)
        with aioresponses() as m:
            mocked_chunks = [b"Cavefish ", b"have ", b"no ", b"sight."]
            mocked_response = b"".join(mocked_chunks)
            m.get("http://example.com", status=200, body=mocked_response)
            response = await request("http://example.com")
            assert response.status_code == 200
            assert response.headers == {"Content-Type": "application/json"}
            content = b"".join([chunk async for chunk in response.content()])
            assert content == b"Cavefish have no sight."

    async def test_request_call_raises_client_error(self, aiohttp_request):
        with aioresponses() as m:
            m.get("http://example.com", exception=aiohttp.ClientError)

            with pytest.raises(exceptions.TransportError) as exc:
                await aiohttp_request("http://example.com/api")

            exc.match("Failed to send request to http://example.com/api.")

    async def test_request_call_raises_timeout_error(self, aiohttp_request):
        with aioresponses() as m:
            m.get("http://example.com", exception=asyncio.TimeoutError)

            with pytest.raises(exceptions.TimeoutError) as exc:
                await aiohttp_request(
                    "http://example.com", timeout=aiohttp.ClientTimeout(total=120)
                )

            exc.match("Request timed out after 120 seconds.")

    async def test_request_call_raises_transport_error_for_closed_session(
        self, aiohttp_request
    ):
        with aioresponses() as m:
            m.get("http://example.com", exception=asyncio.TimeoutError)
            aiohttp_request._closed = True
            with pytest.raises(exceptions.TransportError) as exc:
                await aiohttp_request("http://example.com")

            exc.match("session is closed.")
            aiohttp_request._closed = False

    async def test_request_clone(self):
        request = auth_aiohttp.Request()
        cloned = request._clone()
        assert cloned is not request
        assert isinstance(cloned, auth_aiohttp.Request)
        assert cloned._session is not request._session
        await request.close()
        await cloned.close()

    async def test_request_close(self):
        request = auth_aiohttp.Request()
        assert not getattr(request, "_closed", False)
        await request.close()
        assert request._closed
        # Second call should be idempotent
        await request.close()
        assert request._closed

    async def test_request_clone_closed_session_raises(self):
        request = auth_aiohttp.Request()
        await request.close()
        with pytest.raises(exceptions.TransportError) as exc:
            request._clone()
        exc.match("Cannot clone a closed transport.")

    async def test_request_clone_with_active_session(self):
        import ssl
        from aiohttp import BasicAuth, ClientTimeout, TCPConnector

        custom_ssl = ssl.create_default_context()
        custom_connector = TCPConnector(
            ssl=custom_ssl,
            limit=42,
            limit_per_host=12,
            force_close=True,
            local_addr=("127.0.0.2", 0),
        )

        mock_session = aiohttp.ClientSession(
            connector=custom_connector,
            headers={"x-corporate-firewall": "open"},
            cookies={"enterprise_session": "active"},
            auth=BasicAuth("admin", "secret"),
            timeout=ClientTimeout(total=84.0),
            trust_env=True,
            trace_configs=[aiohttp.TraceConfig()],
        )
        request = auth_aiohttp.Request(session=mock_session)

        cloned = request._clone()

        assert cloned is not request
        assert cloned._session is not mock_session
        assert cloned._session is not None

        # Verify underlying TCPConnector configuration
        cloned_connector = cloned._session._connector
        assert isinstance(cloned_connector, TCPConnector)
        assert cloned_connector is not custom_connector
        assert cloned_connector._resolver is not custom_connector._resolver
        assert cloned_connector._ssl is custom_ssl
        assert cloned_connector._limit == 42
        assert cloned_connector._limit_per_host == 12
        assert cloned_connector._force_close is True
        assert _helpers_async._get_local_addr(cloned_connector) == (
            "127.0.0.2",
            0,
        )

        # Verify session-level configuration
        assert cloned._session._trust_env is True
        assert len(cloned._session._trace_configs) == 1
        assert cloned._session._default_headers == {"x-corporate-firewall": "open"}
        assert cloned._session._cookie_jar is mock_session._cookie_jar
        assert cloned._session._default_auth == mock_session._default_auth
        assert cloned._session._timeout == ClientTimeout(total=84.0)

        await request.close()
        await cloned.close()

    async def test_request_clone_unix_socket(self):
        try:
            from aiohttp import UnixConnector
        except ImportError:
            return  # Windows or environment without Unix Domain Sockets

        connector = UnixConnector(path="/var/run/enterprise.sock", limit=42)
        mock_session = aiohttp.ClientSession(connector=connector)
        request = auth_aiohttp.Request(session=mock_session)

        cloned = request._clone()

        assert cloned._session is not None
        cloned_connector = cloned._session._connector
        assert isinstance(cloned_connector, UnixConnector)
        assert cloned_connector._path == "/var/run/enterprise.sock"
        assert cloned_connector._limit == 42

        await request.close()
        await cloned.close()

    async def test_request_call_raises_timeout_error_int(self, aiohttp_request):
        with aioresponses() as m:
            m.get("http://example.com", exception=asyncio.TimeoutError)
            with pytest.raises(exceptions.TimeoutError) as exc:
                await aiohttp_request("http://example.com", timeout=120)
            exc.match("Request timed out after 120 seconds.")

    async def test_request_clone_with_closed_connector(self):
        session = aiohttp.ClientSession()
        request = auth_aiohttp.Request(session=session)
        await session.close()

        cloned = request._clone()
        assert cloned is not request
        assert cloned._session is not None
        await request.close()
        await cloned.close()

    async def test_request_clone_with_custom_connector(self):
        session = aiohttp.ClientSession()
        custom_connector = AsyncMock()
        custom_connector.closed = False
        custom_connector.close = AsyncMock()
        session._connector = custom_connector

        request = auth_aiohttp.Request(session=session)
        with pytest.raises(
            exceptions.TransportError, match="Unsupported connector type for cloning"
        ):
            request._clone()
        await request.close()

    async def test_request_clone_unix_socket_no_path(self):
        try:
            from aiohttp import UnixConnector
        except ImportError:
            return

        session = aiohttp.ClientSession()
        connector = UnixConnector(path="/tmp/test.sock")
        connector._path = None
        session._connector = connector

        request = auth_aiohttp.Request(session=session)
        cloned = request._clone()
        assert cloned is not request
        assert cloned._session is not None
        assert cloned._session._connector is not connector
        await request.close()
        await cloned.close()



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import ssl
from unittest import mock

import pytest

from google.auth import exceptions
from google.auth.aio import credentials
from google.auth.aio import transport
from google.auth.aio.transport import sessions

# This is the valid "workload" format the library expects
VALID_WORKLOAD_CONFIG = {
    "version": 1,
    "cert_configs": {
        "workload": {"cert_path": "/tmp/mock_cert.pem", "key_path": "/tmp/mock_key.pem"}
    },
}


class TestSessionsMtls:
    @pytest.mark.asyncio
    async def test_configure_mtls_channel(self):
        """
        Tests that the mTLS channel configures correctly when a
        valid workload config is mocked.
        """
        with mock.patch.dict(
            os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}
        ), mock.patch("os.path.exists") as mock_exists, mock.patch(
            "builtins.open", mock.mock_open(read_data=json.dumps(VALID_WORKLOAD_CONFIG))
        ), mock.patch(
            "google.auth.aio.transport.mtls.get_client_cert_and_key"
        ) as mock_helper, mock.patch(
            "google.auth.aio.transport.mtls.make_client_cert_ssl_context"
        ) as mock_make_context, mock.patch(
            "aiohttp.TCPConnector"
        ) as mock_connector, mock.patch(
            "aiohttp.ClientSession"
        ) as mock_session:
            mock_session.return_value.close = mock.AsyncMock()
            mock_exists.return_value = True
            mock_helper.return_value = (True, b"fake_cert_data", b"fake_key_data")

            mock_context = mock.Mock(spec=ssl.SSLContext)
            mock_make_context.return_value = mock_context

            mock_creds = mock.AsyncMock(spec=credentials.Credentials)
            session = sessions.AsyncAuthorizedSession(mock_creds)

            await session.configure_mtls_channel()

            assert session._is_mtls is True
            mock_make_context.assert_called_once_with(
                b"fake_cert_data", b"fake_key_data"
            )
            mock_connector.assert_called_once_with(ssl=mock_context)
            mock_session.assert_called_once_with(connector=mock_connector.return_value)
            await session.close()

    @pytest.mark.asyncio
    async def test_configure_mtls_channel_disabled(self):
        """
        Tests behavior when the config file does not exist.
        """
        with mock.patch.dict(
            os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}
        ), mock.patch("os.path.exists") as mock_exists:
            mock_exists.return_value = False
            mock_creds = mock.AsyncMock(spec=credentials.Credentials)
            session = sessions.AsyncAuthorizedSession(mock_creds)
            await session.configure_mtls_channel()
            assert session._is_mtls is False
            await session.close()

    @pytest.mark.asyncio
    async def test_configure_mtls_channel_invalid_format(self):
        """
        Verifies that the MutualTLSChannelError is raised for bad formats.
        """
        with mock.patch.dict(
            os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}
        ), mock.patch("os.path.exists") as mock_exists, mock.patch(
            "builtins.open", mock.mock_open(read_data='{"invalid": "format"}')
        ):
            mock_exists.return_value = True
            mock_creds = mock.AsyncMock(spec=credentials.Credentials)
            session = sessions.AsyncAuthorizedSession(mock_creds)

            with pytest.raises(exceptions.MutualTLSChannelError):
                await session.configure_mtls_channel()
            await session.close()

    @pytest.mark.asyncio
    async def test_configure_mtls_channel_invalud_fields(self):
        """
        If cert is missing expected keys, it should fail gracefully
        """
        with mock.patch.dict(
            os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}
        ), mock.patch("os.path.exists") as mock_exists, mock.patch(
            "builtins.open", mock.mock_open(read_data='{"cert_configs": {}}')
        ):
            mock_exists.return_value = True
            mock_creds = mock.AsyncMock(spec=credentials.Credentials)
            session = sessions.AsyncAuthorizedSession(mock_creds)
            await session.configure_mtls_channel()
            assert session._is_mtls is False
            await session.close()

    @pytest.mark.asyncio
    async def test_configure_mtls_channel_mock_callback(self):
        """
        Tests mTLS configuration using bytes-returning callback.
        """

        def mock_callback():
            return (b"fake_cert_bytes", b"fake_key_bytes")

        with mock.patch.dict(
            os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}
        ), mock.patch(
            "google.auth.transport.mtls.has_default_client_cert_source",
            return_value=True,
        ), mock.patch(
            "google.auth.aio.transport.mtls.make_client_cert_ssl_context"
        ) as mock_make_context, mock.patch(
            "aiohttp.TCPConnector"
        ) as mock_connector, mock.patch(
            "aiohttp.ClientSession"
        ) as mock_session:
            mock_session.return_value.close = mock.AsyncMock()
            mock_context = mock.Mock(spec=ssl.SSLContext)
            mock_make_context.return_value = mock_context

            mock_creds = mock.AsyncMock(spec=credentials.Credentials)
            session = sessions.AsyncAuthorizedSession(mock_creds)

            await session.configure_mtls_channel(client_cert_callback=mock_callback)

            assert session._is_mtls is True
            mock_make_context.assert_called_once_with(
                b"fake_cert_bytes", b"fake_key_bytes"
            )
            mock_connector.assert_called_once_with(ssl=mock_context)
            mock_session.assert_called_once_with(connector=mock_connector.return_value)
            await session.close()

    @pytest.mark.asyncio
    async def test_configure_mtls_channel_custom_request(self):
        """Tests that if _auth_request is not an AiohttpRequest, _is_mtls is set to False
        because we can't configure the custom request with mTLS.
        """
        with mock.patch.dict(
            os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}
        ), mock.patch("os.path.exists") as mock_exists, mock.patch(
            "builtins.open", mock.mock_open(read_data=json.dumps(VALID_WORKLOAD_CONFIG))
        ), mock.patch(
            "google.auth.aio.transport.mtls.get_client_cert_and_key"
        ) as mock_helper, mock.patch(
            "google.auth.aio.transport.mtls.make_client_cert_ssl_context"
        ) as mock_make_context:
            mock_exists.return_value = True
            mock_helper.return_value = (True, b"fake_cert_data", b"fake_key_data")

            mock_context = mock.Mock(spec=ssl.SSLContext)
            mock_make_context.return_value = mock_context

            mock_creds = mock.AsyncMock(spec=credentials.Credentials)
            mock_auth_request = mock.AsyncMock(spec=transport.Request)
            session = sessions.AsyncAuthorizedSession(
                mock_creds, auth_request=mock_auth_request
            )

            with pytest.warns(UserWarning, match="Attempted to establish mTLS"):
                await session.configure_mtls_channel()

            # If the request handler is not an AiohttpRequest, the library cannot configure
            # the connection to use mTLS, so _is_mtls must be False to reflect this unconfigured state.
            assert session._is_mtls is False
            mock_make_context.assert_called_once_with(
                b"fake_cert_data", b"fake_key_data"
            )
            await session.close()

    @pytest.mark.asyncio
    async def test_configure_mtls_channel_exception_resets_flag(self):
        """
        Tests that self._is_mtls is reset to False if an exception is raised
        during configuration.
        """
        with mock.patch.dict(
            os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}
        ), mock.patch("os.path.exists") as mock_exists, mock.patch(
            "builtins.open", mock.mock_open(read_data=json.dumps(VALID_WORKLOAD_CONFIG))
        ), mock.patch(
            "google.auth.aio.transport.mtls.get_client_cert_and_key"
        ) as mock_helper, mock.patch(
            "google.auth.aio.transport.mtls.make_client_cert_ssl_context"
        ) as mock_make_context:
            mock_exists.return_value = True
            mock_helper.return_value = (True, b"fake_cert_data", b"fake_key_data")
            mock_make_context.side_effect = exceptions.ClientCertError("Mock error")

            mock_creds = mock.AsyncMock(spec=credentials.Credentials)
            session = sessions.AsyncAuthorizedSession(mock_creds)

            with pytest.raises(exceptions.MutualTLSChannelError):
                await session.configure_mtls_channel()

            assert session._is_mtls is False
            await session.close()

    @pytest.mark.asyncio
    async def test_configure_mtls_channel_transport_error_resets_flag(self):
        """
        Tests that self._is_mtls is reset to False if a TransportError is raised
        during configuration.
        """
        with mock.patch.dict(
            os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}
        ), mock.patch("os.path.exists") as mock_exists, mock.patch(
            "builtins.open", mock.mock_open(read_data=json.dumps(VALID_WORKLOAD_CONFIG))
        ), mock.patch(
            "google.auth.aio.transport.mtls.get_client_cert_and_key"
        ) as mock_helper, mock.patch(
            "google.auth.aio.transport.mtls.make_client_cert_ssl_context"
        ) as mock_make_context:
            mock_exists.return_value = True
            mock_helper.return_value = (True, b"fake_cert_data", b"fake_key_data")
            mock_make_context.side_effect = exceptions.TransportError("Mock error")

            mock_creds = mock.AsyncMock(spec=credentials.Credentials)
            session = sessions.AsyncAuthorizedSession(mock_creds)

            with pytest.raises(exceptions.MutualTLSChannelError):
                await session.configure_mtls_channel()

            assert session._is_mtls is False
            await session.close()

    @pytest.mark.asyncio
    async def test_configure_mtls_channel_atomic_on_exception(self):
        """
        Tests that if configure_mtls_channel has already successfully configured mTLS,
        a subsequent attempt that raises an exception will preserve the original mTLS state.
        """
        # Step 1: Successful configuration
        with mock.patch.dict(
            os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}
        ), mock.patch("os.path.exists") as mock_exists, mock.patch(
            "builtins.open", mock.mock_open(read_data=json.dumps(VALID_WORKLOAD_CONFIG))
        ), mock.patch(
            "google.auth.aio.transport.mtls.get_client_cert_and_key"
        ) as mock_helper, mock.patch(
            "google.auth.aio.transport.mtls.make_client_cert_ssl_context"
        ) as mock_make_context, mock.patch(
            "aiohttp.TCPConnector"
        ), mock.patch(
            "aiohttp.ClientSession"
        ) as mock_session:
            mock_session.return_value.close = mock.AsyncMock()
            mock_exists.return_value = True
            mock_helper.return_value = (True, b"fake_cert_data_1", b"fake_key_data_1")

            mock_context = mock.Mock(spec=ssl.SSLContext)
            mock_make_context.return_value = mock_context

            mock_creds = mock.AsyncMock(spec=credentials.Credentials)
            session = sessions.AsyncAuthorizedSession(mock_creds)

            await session.configure_mtls_channel()
            assert session._is_mtls is True
            assert session._cached_cert == b"fake_cert_data_1"
            first_auth_request = session._auth_request

            # Step 2: Failed subsequent configuration attempt
            # Reset task so we trigger a new configuration run
            session._mtls_init_task = None

            # Patch context generator to fail this time
            mock_make_context.side_effect = exceptions.ClientCertError("Mock error")

            with pytest.raises(exceptions.MutualTLSChannelError):
                await session.configure_mtls_channel()

            # Verify that the state remains unchanged from the first successful configuration
            assert session._is_mtls is True
            assert session._cached_cert == b"fake_cert_data_1"
            assert session._auth_request is first_auth_request
            await session.close()

    @pytest.mark.asyncio
    async def test_configure_mtls_channel_close_exception_does_not_abort(self):
        """
        Tests that if old_auth_request.close() raises an exception, the mTLS
        configuration is still considered successful, and is_mtls remains True
        without raising MutualTLSChannelError.
        """
        with mock.patch.dict(
            os.environ, {"GOOGLE_API_USE_CLIENT_CERTIFICATE": "true"}
        ), mock.patch("os.path.exists") as mock_exists, mock.patch(
            "builtins.open", mock.mock_open(read_data=json.dumps(VALID_WORKLOAD_CONFIG))
        ), mock.patch(
            "google.auth.aio.transport.mtls.get_client_cert_and_key"
        ) as mock_helper, mock.patch(
            "google.auth.aio.transport.mtls.make_client_cert_ssl_context"
        ) as mock_make_context, mock.patch(
            "aiohttp.TCPConnector"
        ), mock.patch(
            "aiohttp.ClientSession"
        ) as mock_session:
            mock_session.return_value.close = mock.AsyncMock()
            mock_exists.return_value = True
            mock_helper.return_value = (True, b"fake_cert_data", b"fake_key_data")

            mock_context = mock.Mock(spec=ssl.SSLContext)
            mock_make_context.return_value = mock_context

            mock_creds = mock.AsyncMock(spec=credentials.Credentials)
            session = sessions.AsyncAuthorizedSession(mock_creds)

            # Mock close() of the initial self._auth_request to raise an exception
            session._auth_request.close = mock.AsyncMock(
                side_effect=Exception("Mock close error")
            )

            # Should complete successfully without raising MutualTLSChannelError
            await session.configure_mtls_channel()

            assert session._is_mtls is True
            assert session._cached_cert == b"fake_cert_data"
            await session.close()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import ssl
from unittest import mock

import pytest  # type: ignore

from google.auth import exceptions
from google.auth.transport import _mtls_helper
from google.auth.transport import mtls


@mock.patch("google.auth.transport._mtls_helper._check_config_path")
def test_has_default_client_cert_source_with_context_aware_metadata(mock_check):
    """
    Directly tests the logic: if CONTEXT_AWARE_METADATA_PATH is found, return True.
    """

    # Setup: Return a path only for the Context Aware Metadata Path
    def side_effect(path):
        if path == _mtls_helper.CONTEXT_AWARE_METADATA_PATH:
            return "/path/to/context_aware_metadata.json"
        return None

    mock_check.side_effect = side_effect

    # Execute
    result = mtls.has_default_client_cert_source(True)

    # Assert
    assert result is True
    mock_check.assert_any_call(_mtls_helper.CONTEXT_AWARE_METADATA_PATH)
    assert side_effect("non-matching-path") is None


@mock.patch("google.auth.transport._mtls_helper._check_config_path")
def test_has_default_client_cert_source_falls_back(mock_check):
    """
    Tests that it skips CONTEXT_AWARE_METADATA_PATH if None, and checks the next path.
    """

    # Setup: First path is None, second path is valid
    def side_effect(path):
        if path == _mtls_helper.CERTIFICATE_CONFIGURATION_DEFAULT_PATH:
            return "/path/to/default_cert.json"
        return None

    mock_check.side_effect = side_effect

    # Execute
    result = mtls.has_default_client_cert_source(True)

    # Assert
    assert result is True
    # Verify the sequence of calls
    expected_calls = [
        mock.call(_mtls_helper.CONTEXT_AWARE_METADATA_PATH),
        mock.call(_mtls_helper.CERTIFICATE_CONFIGURATION_DEFAULT_PATH),
    ]
    mock_check.assert_has_calls(expected_calls)


@mock.patch("google.auth.transport.mtls.getenv", autospec=True)
@mock.patch("google.auth.transport._mtls_helper._check_config_path", autospec=True)
def test_has_default_client_cert_source_env_var_success(check_config_path, mock_getenv):
    # 1. Mock getenv to return our test path
    mock_getenv.side_effect = lambda var: (
        "path/to/cert.json" if var == "GOOGLE_API_CERTIFICATE_CONFIG" else None
    )

    # 2. Mock _check_config_path side effect
    def side_effect(path):
        # Return None for legacy paths to ensure we reach the env var logic
        if path == "path/to/cert.json":
            return "/absolute/path/to/cert.json"
        return None

    check_config_path.side_effect = side_effect

    # 3. This should now return True
    assert mtls.has_default_client_cert_source(True)

    # 4. Verify the env var path was checked
    check_config_path.assert_called_with("path/to/cert.json")


@mock.patch("google.auth.transport.mtls.getenv", autospec=True)
@mock.patch("google.auth.transport._mtls_helper._check_config_path", autospec=True)
def test_has_default_client_cert_source_env_var_invalid_config_path(
    check_config_path, mock_getenv
):
    # Set the env var but make the check fail
    mock_getenv.side_effect = lambda var: (
        "invalid/path" if var == "GOOGLE_API_CERTIFICATE_CONFIG" else None
    )
    check_config_path.return_value = None

    assert not mtls.has_default_client_cert_source(True)


@mock.patch("google.auth.transport._mtls_helper.get_client_cert_and_key", autospec=True)
@mock.patch("google.auth.transport.mtls.has_default_client_cert_source", autospec=True)
def test_default_client_cert_source(
    has_default_client_cert_source, get_client_cert_and_key
):
    # Test default client cert source doesn't exist.
    has_default_client_cert_source.return_value = False
    with pytest.raises(exceptions.MutualTLSChannelError):
        mtls.default_client_cert_source()

    # The following tests will assume default client cert source exists.
    has_default_client_cert_source.return_value = True

    # Test good callback.
    get_client_cert_and_key.return_value = (True, b"cert", b"key")
    callback = mtls.default_client_cert_source()
    assert callback() == (b"cert", b"key")

    # Test bad callback which throws exception.
    get_client_cert_and_key.side_effect = ValueError()
    callback = mtls.default_client_cert_source()
    with pytest.raises(exceptions.MutualTLSChannelError):
        callback()

    # Test bad callback which throws ClientCertError.
    get_client_cert_and_key.side_effect = exceptions.ClientCertError()
    callback = mtls.default_client_cert_source()
    with pytest.raises(exceptions.ClientCertError):
        callback()


@mock.patch(
    "google.auth.transport._mtls_helper.get_client_ssl_credentials", autospec=True
)
@mock.patch("google.auth.transport.mtls.has_default_client_cert_source", autospec=True)
def test_default_client_encrypted_cert_source(
    has_default_client_cert_source, get_client_ssl_credentials
):
    # Test default client cert source doesn't exist.
    has_default_client_cert_source.return_value = False
    with pytest.raises(exceptions.MutualTLSChannelError):
        mtls.default_client_encrypted_cert_source("cert_path", "key_path")

    # The following tests will assume default client cert source exists.
    has_default_client_cert_source.return_value = True

    # Test good callback.
    get_client_ssl_credentials.return_value = (True, b"cert", b"key", b"passphrase")
    callback = mtls.default_client_encrypted_cert_source("cert_path", "key_path")
    with mock.patch("google.auth.transport.mtls.open", mock.mock_open()) as mock_file:
        assert callback() == ("cert_path", "key_path", b"passphrase")
        mock_file.assert_any_call("cert_path", "wb")
        mock_file.assert_any_call("key_path", "wb")

    # Test bad callback which throws exception.
    get_client_ssl_credentials.side_effect = exceptions.ClientCertError()
    callback = mtls.default_client_encrypted_cert_source("cert_path", "key_path")
    with pytest.raises(exceptions.MutualTLSChannelError):
        callback()


@mock.patch("google.auth.transport._mtls_helper.check_use_client_cert", autospec=True)
def test_should_use_client_cert(check_use_client_cert):
    check_use_client_cert.return_value = mock.Mock()
    assert mtls.should_use_client_cert()

    check_use_client_cert.return_value = False
    assert not mtls.should_use_client_cert()


@contextlib.contextmanager
def _fake_secure_paths(cert_bytes, key_bytes, passphrase=None):
    yield "cert_path", "key_path", passphrase


@mock.patch(
    "google.auth.transport._mtls_helper.secure_cert_key_paths",
    side_effect=_fake_secure_paths,
)
def test_load_client_cert_into_context_success(mock_secure_paths):
    mock_ctx = mock.Mock(spec=ssl.SSLContext)
    result = mtls._load_client_cert_into_context(
        mock_ctx, b"cert", b"key", passphrase=b"passphrase"
    )
    assert result is None
    mock_secure_paths.assert_called_once_with(b"cert", b"key", passphrase=b"passphrase")
    mock_ctx.load_cert_chain.assert_called_once_with(
        certfile="cert_path", keyfile="key_path", password=b"passphrase"
    )


@mock.patch(
    "google.auth.transport._mtls_helper.secure_cert_key_paths",
    side_effect=_fake_secure_paths,
)
def test_load_client_cert_into_context_success_no_passphrase(mock_secure_paths):
    mock_ctx = mock.Mock(spec=ssl.SSLContext)
    result = mtls._load_client_cert_into_context(mock_ctx, b"cert", b"key")
    assert result is None
    mock_secure_paths.assert_called_once_with(b"cert", b"key", passphrase=None)
    mock_ctx.load_cert_chain.assert_called_once_with(
        certfile="cert_path", keyfile="key_path", password=None
    )


@pytest.mark.parametrize(
    "exception",
    [
        ssl.SSLError("mock error message"),
        OSError("mock error message"),
        ValueError("mock error message"),
        RuntimeError("mock error message"),
        TypeError("mock error message"),
    ],
)
@mock.patch(
    "google.auth.transport._mtls_helper.secure_cert_key_paths",
    side_effect=_fake_secure_paths,
)
def test_load_client_cert_into_context_error(mock_secure_paths, exception):
    mock_ctx = mock.Mock(spec=ssl.SSLContext)
    mock_ctx.load_cert_chain.side_effect = exception
    with pytest.raises(exceptions.MutualTLSChannelError) as exc_info:
        mtls._load_client_cert_into_context(mock_ctx, b"cert", b"key")
    assert "mock error message" in str(exc_info.value)
    assert isinstance(exc_info.value.__cause__, type(exception))


@pytest.mark.parametrize("invalid_ctx", [None, object()])
def test_load_client_cert_into_context_invalid_ctx(invalid_ctx):
    with pytest.raises(exceptions.MutualTLSChannelError) as exc_info:
        mtls._load_client_cert_into_context(invalid_ctx, b"cert", b"key")
    assert (
        "The provided context object is invalid or does not support loading certificate chains"
        in str(exc_info.value)
    )
    assert exc_info.value.__cause__ is None


@mock.patch("google.auth.transport.mtls.should_use_client_cert", autospec=True)
def test_load_default_client_cert_disabled(mock_should_use):
    mock_should_use.return_value = False
    mock_ctx = mock.Mock(spec=ssl.SSLContext)
    assert mtls.load_default_client_cert(mock_ctx) is False
    mock_ctx.load_cert_chain.assert_not_called()


@mock.patch("google.auth.transport.mtls.has_default_client_cert_source", autospec=True)
@mock.patch("google.auth.transport.mtls.should_use_client_cert", autospec=True)
def test_load_default_client_cert_no_source(mock_should_use, mock_has_source):
    mock_should_use.return_value = True
    mock_has_source.return_value = False
    mock_ctx = mock.Mock(spec=ssl.SSLContext)
    assert mtls.load_default_client_cert(mock_ctx) is False
    mock_ctx.load_cert_chain.assert_not_called()


@mock.patch(
    "google.auth.transport._mtls_helper.get_client_ssl_credentials", autospec=True
)
@mock.patch("google.auth.transport.mtls.has_default_client_cert_source", autospec=True)
@mock.patch("google.auth.transport.mtls.should_use_client_cert", autospec=True)
def test_load_default_client_cert_no_cert(
    mock_should_use, mock_has_source, mock_get_credentials
):
    mock_should_use.return_value = True
    mock_has_source.return_value = True
    mock_get_credentials.return_value = (False, None, None, None)
    mock_ctx = mock.Mock(spec=ssl.SSLContext)
    assert mtls.load_default_client_cert(mock_ctx) is False
    mock_ctx.load_cert_chain.assert_not_called()


@mock.patch(
    "google.auth.transport._mtls_helper.secure_cert_key_paths",
    side_effect=_fake_secure_paths,
)
@mock.patch(
    "google.auth.transport._mtls_helper.get_client_ssl_credentials", autospec=True
)
@mock.patch("google.auth.transport.mtls.has_default_client_cert_source", autospec=True)
@mock.patch("google.auth.transport.mtls.should_use_client_cert", autospec=True)
def test_load_default_client_cert_success(
    mock_should_use, mock_has_source, mock_get_credentials, mock_secure_paths
):
    mock_should_use.return_value = True
    mock_has_source.return_value = True
    mock_get_credentials.return_value = (True, b"cert", b"key", b"passphrase")
    mock_ctx = mock.Mock(spec=ssl.SSLContext)

    assert mtls.load_default_client_cert(mock_ctx) is True
    mock_ctx.load_cert_chain.assert_called_once_with(
        certfile="cert_path", keyfile="key_path", password=b"passphrase"
    )


@mock.patch(
    "google.auth.transport._mtls_helper.get_client_ssl_credentials", autospec=True
)
@mock.patch("google.auth.transport.mtls.has_default_client_cert_source", autospec=True)
@mock.patch("google.auth.transport.mtls.should_use_client_cert", autospec=True)
def test_load_default_client_cert_propagates_client_cert_error(
    mock_should_use, mock_has_source, mock_get_credentials
):
    mock_should_use.return_value = True
    mock_has_source.return_value = True
    mock_get_credentials.side_effect = exceptions.ClientCertError("credentials failure")
    mock_ctx = mock.Mock(spec=ssl.SSLContext)

    with pytest.raises(exceptions.MutualTLSChannelError) as exc_info:
        mtls.load_default_client_cert(mock_ctx)
    assert "credentials failure" in str(exc_info.value)
    assert isinstance(exc_info.value.__cause__, exceptions.ClientCertError)


@mock.patch("google.auth.transport.mtls.has_default_client_cert_source", autospec=True)
@mock.patch("google.auth.transport.mtls.should_use_client_cert", autospec=True)
@mock.patch("google.auth.transport.mtls.load_default_client_cert", autospec=True)
@mock.patch("ssl.create_default_context", autospec=True)
def test_get_default_ssl_context_configured(
    mock_create_context, mock_load_default, mock_should_use, mock_has_source
):
    mock_should_use.return_value = True
    mock_has_source.return_value = True
    mock_ctx = mock.Mock(spec=ssl.SSLContext)
    mock_create_context.return_value = mock_ctx
    mock_load_default.return_value = True

    result = mtls.get_default_ssl_context()

    assert result == mock_ctx
    mock_create_context.assert_called_once_with(ssl.Purpose.SERVER_AUTH)
    mock_load_default.assert_called_once_with(mock_ctx)


@mock.patch("google.auth.transport.mtls.has_default_client_cert_source", autospec=True)
@mock.patch("google.auth.transport.mtls.should_use_client_cert", autospec=True)
@mock.patch("google.auth.transport.mtls.load_default_client_cert", autospec=True)
@mock.patch("ssl.create_default_context", autospec=True)
def test_get_default_ssl_context_unconfigured(
    mock_create_context, mock_load_default, mock_should_use, mock_has_source
):
    mock_should_use.return_value = True
    mock_has_source.return_value = True
    mock_ctx = mock.Mock(spec=ssl.SSLContext)
    mock_create_context.return_value = mock_ctx
    mock_load_default.return_value = False

    result = mtls.get_default_ssl_context()

    assert result is None
    mock_create_context.assert_called_once_with(ssl.Purpose.SERVER_AUTH)
    mock_load_default.assert_called_once_with(mock_ctx)


@mock.patch("google.auth.transport.mtls.has_default_client_cert_source", autospec=True)
@mock.patch("google.auth.transport.mtls.should_use_client_cert", autospec=True)
@mock.patch("google.auth.transport.mtls.load_default_client_cert", autospec=True)
@mock.patch("ssl.create_default_context", autospec=True)
def test_get_default_ssl_context_exception(
    mock_create_context, mock_load_default, mock_should_use, mock_has_source
):
    mock_should_use.return_value = True
    mock_has_source.return_value = True
    mock_ctx = mock.Mock(spec=ssl.SSLContext)
    mock_create_context.return_value = mock_ctx
    mock_load_default.side_effect = exceptions.MutualTLSChannelError(
        "mock error message"
    )

    with pytest.raises(exceptions.MutualTLSChannelError) as exc_info:
        mtls.get_default_ssl_context()
    assert "mock error message" in str(exc_info.value)


@pytest.mark.parametrize(
    "env_val,client_cert_available,expected",
    [
        ("always", True, True),
        ("always", False, True),
        ("never", True, False),
        ("never", False, False),
        ("auto", True, True),
        ("auto", False, False),
        (None, True, True),  # Defaults to auto
        (None, False, False),  # Defaults to auto
        ("ALWAYS", True, True),
        ("ALWAYS", False, True),
        ("NEVER", True, False),
        ("NEVER", False, False),
        ("AUTO", True, True),
        ("AUTO", False, False),
        ("", True, True),
        ("", False, False),
    ],
)
@mock.patch("google.auth.transport.mtls.getenv", autospec=True)
def test_should_use_mtls_endpoint(
    mock_getenv, env_val, client_cert_available, expected
):
    mock_getenv.side_effect = lambda var, default=None: (
        env_val
        if (var == "GOOGLE_API_USE_MTLS_ENDPOINT" and env_val is not None)
        else default
    )
    result = mtls.should_use_mtls_endpoint(client_cert_available)
    assert result == expected


@pytest.mark.parametrize(
    "invalid_value",
    [
        "invalid_value",
        "   ",
        "\t",
    ],
)
@mock.patch("google.auth.transport.mtls.getenv", autospec=True)
def test_should_use_mtls_endpoint_invalid_value(mock_getenv, invalid_value):
    mock_getenv.side_effect = lambda var, default=None: (
        invalid_value if var == "GOOGLE_API_USE_MTLS_ENDPOINT" else default
    )
    with pytest.raises(exceptions.MutualTLSChannelError) as exc_info:
        mtls.should_use_mtls_endpoint(True)
    assert "Unsupported GOOGLE_API_USE_MTLS_ENDPOINT value" in str(exc_info.value)
    assert f"'{invalid_value.strip().lower()}'" in str(exc_info.value)


@mock.patch("google.auth.transport.mtls.should_use_client_cert", autospec=True)
@mock.patch("google.auth.transport.mtls.getenv", autospec=True)
def test_should_use_mtls_endpoint_default_client_cert(
    mock_getenv, mock_should_use_client_cert
):
    mock_getenv.side_effect = lambda var, default=None: (
        "auto" if var == "GOOGLE_API_USE_MTLS_ENDPOINT" else default
    )
    mock_should_use_client_cert.return_value = True
    assert mtls.should_use_mtls_endpoint() is True
    mock_should_use_client_cert.assert_called_once()

    mock_should_use_client_cert.reset_mock()

    mock_should_use_client_cert.return_value = False
    assert mtls.should_use_mtls_endpoint() is False
    mock_should_use_client_cert.assert_called_once()


@mock.patch("google.auth.transport.mtls.should_use_client_cert", autospec=True)
@mock.patch("google.auth.transport.mtls.has_default_client_cert_source", autospec=True)
def test_get_default_ssl_context_not_should_use(mock_has_default, mock_should_use):
    mock_should_use.return_value = False
    mock_has_default.return_value = True

    result = mtls.get_default_ssl_context()
    assert result is None


@mock.patch("google.auth.transport.mtls.should_use_client_cert", autospec=True)
@mock.patch("google.auth.transport.mtls.has_default_client_cert_source", autospec=True)
def test_get_default_ssl_context_no_default_source(mock_has_default, mock_should_use):
    mock_should_use.return_value = True
    mock_has_default.return_value = False

    result = mtls.get_default_ssl_context()
    assert result is None



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

import aiohttp
from aioresponses.core import RequestMatch  # type: ignore


class _DummyWriter:
    output_size = 0

    def done(self):
        return True

    def add_done_callback(self, fn, *args, **kwargs):
        pass

    def remove_done_callback(self, fn):
        pass

    def __await__(self):
        if False:
            yield
        return None


class _CompatClientResponse(aiohttp.ClientResponse):
    """ClientResponse subclass for aioresponses compatibility across all aiohttp versions."""

    def __init__(self, *args, **kwargs):
        writer = kwargs.pop("writer", None)
        stream_writer = kwargs.pop("stream_writer", None)
        writer_obj = stream_writer or writer or _DummyWriter()
        sig = inspect.signature(super().__init__)
        if "stream_writer" in sig.parameters:
            kwargs["stream_writer"] = writer_obj
        if "writer" in sig.parameters:
            kwargs["writer"] = writer_obj
        super().__init__(*args, **kwargs)


def patch_aioresponses():
    """Patch RequestMatch.__init__ to use _CompatClientResponse by default."""
    if getattr(RequestMatch, "_is_compat_patched", False):
        return
    orig_init = RequestMatch.__init__

    def _request_match_init(self, *args, **kwargs):
        if kwargs.get("response_class") is None:
            kwargs["response_class"] = _CompatClientResponse
        orig_init(self, *args, **kwargs)

    RequestMatch.__init__ = _request_match_init
    RequestMatch._is_compat_patched = True



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging
from unittest import mock

import pytest  # type: ignore

from google.auth import _credentials_async
from google.auth import _helpers
from google.auth import _regional_access_boundary_utils
from google.auth import credentials
from google.oauth2 import credentials as oauth2_credentials


@pytest.fixture
def rab_caplog(caplog):
    """Fixture to configure logging capture and ensure propagation for RAB utilities."""

    caplog.set_level(
        logging.DEBUG, logger="google.auth._regional_access_boundary_utils"
    )

    google_logger = logging.getLogger("google")
    original_propagate = google_logger.propagate
    google_logger.propagate = True
    try:
        yield caplog
    finally:
        google_logger.propagate = original_propagate


class CredentialsImpl(credentials.CredentialsWithRegionalAccessBoundary):
    def __init__(self, universe_domain=None):
        super(CredentialsImpl, self).__init__()
        if universe_domain:
            self._universe_domain = universe_domain

    def _perform_refresh_token(self, request):
        self.token = "refreshed-token"
        self.expiry = (
            datetime.datetime.utcnow()
            + _helpers.REFRESH_THRESHOLD
            + datetime.timedelta(seconds=5)
        )

    def with_quota_project(self, quota_project_id):
        raise NotImplementedError()

    def _build_regional_access_boundary_lookup_url(self):
        # Using self.token here to make the URL dynamic for testing purposes
        return "http://mock.url/lookup_for_{}".format(self.token)

    def _make_copy(self):
        new_credentials = self.__class__()
        self._copy_regional_access_boundary_manager(new_credentials)
        return new_credentials


class TestCredentialsWithRegionalAccessBoundary(object):
    @mock.patch(
        "google.auth._regional_access_boundary_utils._RegionalAccessBoundaryRefreshManager.start_refresh"
    )
    def test_maybe_start_refresh_is_skipped_if_not_expired(self, mock_start_refresh):
        creds = CredentialsImpl()
        creds._rab_manager._data = _regional_access_boundary_utils._RegionalAccessBoundaryData(
            encoded_locations="0xABC",
            expiry=_helpers.utcnow() + datetime.timedelta(hours=2),
            cooldown_expiry=None,
            cooldown_duration=_regional_access_boundary_utils.DEFAULT_REGIONAL_ACCESS_BOUNDARY_COOLDOWN,
        )
        creds._maybe_start_regional_access_boundary_refresh(
            mock.Mock(), "http://example.com"
        )
        mock_start_refresh.assert_not_called()

    @mock.patch(
        "google.auth._regional_access_boundary_utils._RegionalAccessBoundaryRefreshManager.start_refresh"
    )
    def test_maybe_start_refresh_triggered_if_soft_expired(self, mock_start_refresh):
        creds = CredentialsImpl()
        creds._rab_manager._data = _regional_access_boundary_utils._RegionalAccessBoundaryData(
            encoded_locations="0xABC",
            expiry=_helpers.utcnow() + datetime.timedelta(minutes=30),
            cooldown_expiry=None,
            cooldown_duration=_regional_access_boundary_utils.DEFAULT_REGIONAL_ACCESS_BOUNDARY_COOLDOWN,
        )
        request = mock.Mock()
        creds._maybe_start_regional_access_boundary_refresh(
            request, "http://example.com"
        )
        mock_start_refresh.assert_called_once_with(creds, request, creds._rab_manager)

    @mock.patch(
        "google.auth._regional_access_boundary_utils._RegionalAccessBoundaryRefreshManager.start_refresh"
    )
    def test_maybe_start_refresh_is_skipped_if_cooldown_active(
        self, mock_start_refresh
    ):
        creds = CredentialsImpl()
        creds._rab_manager._data = _regional_access_boundary_utils._RegionalAccessBoundaryData(
            encoded_locations=None,
            expiry=None,
            cooldown_expiry=_helpers.utcnow() + datetime.timedelta(minutes=5),
            cooldown_duration=_regional_access_boundary_utils.DEFAULT_REGIONAL_ACCESS_BOUNDARY_COOLDOWN,
        )
        creds._maybe_start_regional_access_boundary_refresh(
            mock.Mock(), "http://example.com"
        )
        mock_start_refresh.assert_not_called()

    @mock.patch(
        "google.auth._regional_access_boundary_utils._RegionalAccessBoundaryRefreshManager.start_refresh"
    )
    @pytest.mark.parametrize(
        "url",
        [
            "https://my-service.us-east1.rep.googleapis.com",
            "https://my-service.us-east1.rep.sandbox.googleapis.com",
            "https://my-service.us-east1.rep.mtls.googleapis.com",
            "https://my-service.us-east1.rep.mtls.sandbox.googleapis.com",
        ],
    )
    def test_maybe_start_refresh_is_skipped_for_regional_endpoint(
        self, mock_start_refresh, url
    ):
        creds = CredentialsImpl()
        creds._maybe_start_regional_access_boundary_refresh(mock.Mock(), url)
        mock_start_refresh.assert_not_called()

    @mock.patch(
        "google.auth._regional_access_boundary_utils._RegionalAccessBoundaryRefreshManager.start_refresh"
    )
    def test_maybe_start_refresh_is_triggered(self, mock_start_refresh):
        creds = CredentialsImpl()
        request = mock.Mock()
        creds._maybe_start_regional_access_boundary_refresh(
            request, "http://example.com"
        )
        mock_start_refresh.assert_called_once_with(creds, request, creds._rab_manager)

    def test_apply_headers_success(self):
        creds = CredentialsImpl()
        creds._rab_manager._data = _regional_access_boundary_utils._RegionalAccessBoundaryData(
            encoded_locations="0xABC",
            expiry=_helpers.utcnow() + datetime.timedelta(hours=1),
            cooldown_expiry=None,
            cooldown_duration=_regional_access_boundary_utils.DEFAULT_REGIONAL_ACCESS_BOUNDARY_COOLDOWN,
        )
        headers = {}
        creds._rab_manager.apply_headers(headers)
        assert headers == {"x-allowed-locations": "0xABC"}

    def test_apply_headers_removes_header_if_none(self):
        creds = CredentialsImpl()
        creds._rab_manager._data = _regional_access_boundary_utils._RegionalAccessBoundaryData(
            encoded_locations=None,
            expiry=_helpers.utcnow() + datetime.timedelta(hours=1),
            cooldown_expiry=None,
            cooldown_duration=_regional_access_boundary_utils.DEFAULT_REGIONAL_ACCESS_BOUNDARY_COOLDOWN,
        )
        headers = {"x-allowed-locations": "0xABC"}
        creds._rab_manager.apply_headers(headers)
        assert headers == {}

    def test_apply_headers_removes_header_if_empty(self):
        creds = CredentialsImpl()
        creds._rab_manager._data = _regional_access_boundary_utils._RegionalAccessBoundaryData(
            encoded_locations="",
            expiry=_helpers.utcnow() + datetime.timedelta(hours=1),
            cooldown_expiry=None,
            cooldown_duration=_regional_access_boundary_utils.DEFAULT_REGIONAL_ACCESS_BOUNDARY_COOLDOWN,
        )
        headers = {"x-allowed-locations": "0xABC"}
        creds._rab_manager.apply_headers(headers)
        assert headers == {}

    def test_set_blocking_regional_access_boundary_lookup(self):
        creds = CredentialsImpl()
        assert not creds._rab_manager._use_blocking_regional_access_boundary_lookup

        new_creds = creds._set_blocking_regional_access_boundary_lookup()
        assert new_creds is creds
        assert creds._rab_manager._use_blocking_regional_access_boundary_lookup

    def test_set_regional_access_boundary(self):
        creds = CredentialsImpl()
        seed = {
            "encodedLocations": "0xABC",
            "expiry": _helpers.utcnow() + datetime.timedelta(hours=1),
        }
        new_creds = creds._set_regional_access_boundary(seed)
        assert new_creds is creds
        assert creds._rab_manager._data.encoded_locations == "0xABC"
        assert creds._rab_manager._data.expiry == seed["expiry"]
        assert creds._rab_manager._data.cooldown_expiry is None

    def test_regional_access_boundary_getter(self):
        creds = CredentialsImpl()
        assert creds.regional_access_boundary is None

        seed = {
            "encodedLocations": "0xABC",
            "expiry": _helpers.utcnow() + datetime.timedelta(hours=1),
        }
        creds._set_regional_access_boundary(seed)
        assert creds.regional_access_boundary == "0xABC"

    def test_regional_access_boundary_expiry_getter(self):
        creds = CredentialsImpl()
        assert creds.regional_access_boundary_expiry is None

        seed = {
            "encodedLocations": "0xABC",
            "expiry": _helpers.utcnow() + datetime.timedelta(hours=1),
        }
        creds._set_regional_access_boundary(seed)
        assert creds.regional_access_boundary_expiry == seed["expiry"]

    def test_copy_regional_access_boundary_state(self):
        source_creds = CredentialsImpl()
        snapshot = _regional_access_boundary_utils._RegionalAccessBoundaryData(
            encoded_locations="0xABC",
            expiry=_helpers.utcnow(),
            cooldown_expiry=_helpers.utcnow(),
            cooldown_duration=_regional_access_boundary_utils.DEFAULT_REGIONAL_ACCESS_BOUNDARY_COOLDOWN,
        )
        source_creds._rab_manager._data = snapshot

        target_creds = CredentialsImpl()
        source_creds._copy_regional_access_boundary_manager(target_creds)

        assert target_creds._rab_manager is not source_creds._rab_manager
        assert target_creds._rab_manager._data is source_creds._rab_manager._data

    def test_serialization(self):
        import pickle

        manager = _regional_access_boundary_utils._RegionalAccessBoundaryManager()
        manager._data = _regional_access_boundary_utils._RegionalAccessBoundaryData(
            encoded_locations="0xABC",
            expiry=_helpers.utcnow(),
            cooldown_expiry=_helpers.utcnow(),
            cooldown_duration=_regional_access_boundary_utils.DEFAULT_REGIONAL_ACCESS_BOUNDARY_COOLDOWN,
        )

        pickled = pickle.dumps(manager)
        unpickled = pickle.loads(pickled)

        assert unpickled._data == manager._data
        assert unpickled._update_lock is not None
        assert unpickled.refresh_manager._lock is not None
        assert unpickled.refresh_manager._worker is None

    def test_unpickle_old_credentials_without_rab(self):
        creds = CredentialsImpl()
        old_state = creds.__dict__.copy()
        if "_rab_manager" in old_state:
            del old_state["_rab_manager"]
        if "_use_non_blocking_refresh" in old_state:
            del old_state["_use_non_blocking_refresh"]
        if "_refresh_worker" in old_state:
            del old_state["_refresh_worker"]

        new_instance = CredentialsImpl.__new__(CredentialsImpl)
        new_instance.__setstate__(old_state)

        assert hasattr(new_instance, "_rab_manager")
        assert new_instance._rab_manager is not None
        assert new_instance._use_non_blocking_refresh is False
        assert new_instance._refresh_worker is not None

    @mock.patch(
        "google.auth._regional_access_boundary_utils._RegionalAccessBoundaryRefreshManager.start_refresh"
    )
    def test_maybe_start_refresh_is_skipped_if_non_default_universe_domain(
        self, mock_start_refresh
    ):
        creds = CredentialsImpl(universe_domain="not.googleapis.com")
        creds._maybe_start_regional_access_boundary_refresh(
            mock.Mock(), "http://example.com"
        )
        mock_start_refresh.assert_not_called()

    @mock.patch(
        "google.auth._regional_access_boundary_utils._RegionalAccessBoundaryRefreshManager.start_refresh"
    )
    @mock.patch("urllib.parse.urlparse")
    def test_maybe_start_refresh_handles_url_parse_errors(
        self, mock_urlparse, mock_start_refresh
    ):
        mock_urlparse.side_effect = ValueError("Malformed URL")
        creds = CredentialsImpl()
        request = mock.Mock()
        creds._maybe_start_regional_access_boundary_refresh(
            request, "http://malformed-url"
        )
        mock_start_refresh.assert_called_once_with(creds, request, creds._rab_manager)

    @mock.patch(
        "google.auth._regional_access_boundary_utils._RegionalAccessBoundaryManager.start_blocking_refresh"
    )
    def test_maybe_start_refresh_blocking(self, mock_start_blocking_refresh):
        creds = CredentialsImpl()
        creds._rab_manager._use_blocking_regional_access_boundary_lookup = True
        request = mock.Mock()
        creds._maybe_start_regional_access_boundary_refresh(
            request, "http://example.com"
        )
        mock_start_blocking_refresh.assert_called_once_with(creds, request)

    def test_start_blocking_refresh_success(self):
        creds = CredentialsImpl()
        request = mock.Mock()

        with mock.patch.object(
            creds,
            "_lookup_regional_access_boundary",
            return_value={"encodedLocations": "0xABC"},
        ) as mock_lookup:
            creds._rab_manager.start_blocking_refresh(creds, request)

            mock_lookup.assert_called_once_with(request, fail_fast=True)
            assert creds._rab_manager._data.encoded_locations == "0xABC"

    def test_start_blocking_refresh_failure(self):
        creds = CredentialsImpl()
        request = mock.Mock()

        with mock.patch.object(
            creds, "_lookup_regional_access_boundary", side_effect=Exception("error")
        ) as mock_lookup:
            creds._rab_manager.start_blocking_refresh(creds, request)

            mock_lookup.assert_called_once_with(request, fail_fast=True)
            assert creds._rab_manager._data.encoded_locations is None
            assert creds._rab_manager._data.cooldown_expiry is not None

    def test_start_blocking_refresh_with_async_credentials(self):
        creds = CredentialsImpl()
        request = mock.Mock()

        with mock.patch.object(
            creds,
            "_lookup_regional_access_boundary",
            new_callable=mock.AsyncMock,
        ) as mock_lookup:
            creds._rab_manager.start_blocking_refresh(creds, request)

            mock_lookup.assert_not_called()
            assert creds._rab_manager._data.encoded_locations is None
            assert creds._rab_manager._data.cooldown_expiry is not None

    @mock.patch("copy.deepcopy")
    def test_start_refresh_deepcopy_failure(self, mock_deepcopy):
        mock_deepcopy.side_effect = Exception("deepcopy error")
        creds = CredentialsImpl()
        request = mock.Mock()

        creds._rab_manager.refresh_manager.start_refresh(
            creds, request, creds._rab_manager
        )

        assert creds._rab_manager.refresh_manager._worker is None

    @mock.patch.object(CredentialsImpl, "_lookup_regional_access_boundary")
    def test_lookup_regional_access_boundary_success(self, mock_lookup_rab):
        creds = CredentialsImpl()
        request = mock.Mock()
        rab_manager = _regional_access_boundary_utils._RegionalAccessBoundaryManager()

        mock_lookup_rab.return_value = {
            "locations": ["us-east1"],
            "encodedLocations": "0xABC123",
        }

        worker = _regional_access_boundary_utils._RegionalAccessBoundaryRefreshThread(
            creds, request, rab_manager
        )
        worker.run()

        mock_lookup_rab.assert_called_once_with(request)
        assert rab_manager._data.encoded_locations == "0xABC123"
        assert rab_manager._data.expiry is not None
        assert rab_manager._data.cooldown_expiry is None

    @mock.patch.object(CredentialsImpl, "_lookup_regional_access_boundary")
    def test_lookup_regional_access_boundary_failure(self, mock_lookup_rab, rab_caplog):
        creds = CredentialsImpl()
        request = mock.Mock()
        rab_manager = _regional_access_boundary_utils._RegionalAccessBoundaryManager()

        mock_lookup_rab.return_value = None

        worker = _regional_access_boundary_utils._RegionalAccessBoundaryRefreshThread(
            creds, request, rab_manager
        )
        worker.run()

        mock_lookup_rab.assert_called_once_with(request)
        assert rab_manager._data.encoded_locations is None
        assert rab_manager._data.expiry is None
        assert rab_manager._data.cooldown_expiry is not None

        # RAB failures should be logged at DEBUG level.
        assert any(
            t[1] == logging.DEBUG
            and "Regional Access Boundary lookup failed. Entering cooldown." in t[2]
            for t in rab_caplog.record_tuples
        )

    def test_lookup_regional_access_boundary_null_url(self):
        creds = oauth2_credentials.Credentials(token="token")
        request = mock.Mock()
        result = creds._lookup_regional_access_boundary(request)
        assert result is None
        request.assert_not_called()

    def test_credentials_with_regional_access_boundary_initialization(self):
        creds = CredentialsImpl()
        assert creds._rab_manager._data.encoded_locations is None
        assert creds._rab_manager._data.expiry is None
        assert creds._rab_manager._data.cooldown_expiry is None
        assert creds._rab_manager._data.cooldown_duration == (
            credentials._regional_access_boundary_utils.DEFAULT_REGIONAL_ACCESS_BOUNDARY_COOLDOWN
        )
        assert creds._rab_manager._update_lock is not None

    @mock.patch("google.auth._helpers.utcnow")
    def test_regional_access_boundary_refresh_thread_run_success(self, mock_utcnow):
        mock_now = datetime.datetime(2025, 1, 1, 12, 0, 0)
        mock_utcnow.return_value = mock_now

        creds = mock.Mock()
        request = mock.Mock()
        rab_manager = _regional_access_boundary_utils._RegionalAccessBoundaryManager()

        creds._lookup_regional_access_boundary.return_value = {
            "locations": ["us-east1"],
            "encodedLocations": "0xABC123",
        }

        worker = _regional_access_boundary_utils._RegionalAccessBoundaryRefreshThread(
            creds, request, rab_manager
        )
        worker.run()

        assert rab_manager._data.encoded_locations == "0xABC123"
        expected_expiry = (
            mock_now
            + _regional_access_boundary_utils.DEFAULT_REGIONAL_ACCESS_BOUNDARY_TTL
        )
        assert rab_manager._data.expiry == expected_expiry
        assert rab_manager._data.cooldown_expiry is None

    @mock.patch("google.auth._helpers.utcnow")
    def test_regional_access_boundary_refresh_thread_run_failure(
        self, mock_utcnow, rab_caplog
    ):
        mock_now = datetime.datetime(2025, 1, 1, 12, 0, 0)
        mock_utcnow.return_value = mock_now

        creds = mock.Mock()
        request = mock.Mock()
        rab_manager = _regional_access_boundary_utils._RegionalAccessBoundaryManager()

        initial_cooldown = rab_manager._data.cooldown_duration

        creds._lookup_regional_access_boundary.side_effect = Exception(
            "Network failure"
        )

        worker = _regional_access_boundary_utils._RegionalAccessBoundaryRefreshThread(
            creds, request, rab_manager
        )
        worker.run()

        assert rab_manager._data.encoded_locations is None
        assert rab_manager._data.expiry is None
        expected_cooldown_expiry = mock_now + initial_cooldown
        assert rab_manager._data.cooldown_expiry == expected_cooldown_expiry
        assert rab_manager._data.cooldown_duration == initial_cooldown * 2

        # RAB failures should be logged at DEBUG level.
        assert any(
            t[1] == logging.DEBUG
            and "Asynchronous Regional Access Boundary lookup raised an exception"
            in t[2]
            for t in rab_caplog.record_tuples
        )
        assert any(
            t[1] == logging.DEBUG
            and "Regional Access Boundary lookup failed. Entering cooldown." in t[2]
            for t in rab_caplog.record_tuples
        )

    @mock.patch("google.auth._helpers.utcnow")
    def test_regional_access_boundary_refresh_thread_run_failure_hard_expiry(
        self, mock_utcnow
    ):
        mock_now = datetime.datetime(2025, 1, 1, 12, 0, 0)
        mock_utcnow.return_value = mock_now

        creds = mock.Mock()
        request = mock.Mock()
        rab_manager = _regional_access_boundary_utils._RegionalAccessBoundaryManager()

        # Simulate existing data that has hard expired
        rab_manager._data = _regional_access_boundary_utils._RegionalAccessBoundaryData(
            encoded_locations="0xExpiredToken",
            expiry=mock_now - datetime.timedelta(hours=1),
            cooldown_expiry=None,
            cooldown_duration=_regional_access_boundary_utils.DEFAULT_REGIONAL_ACCESS_BOUNDARY_COOLDOWN,
        )

        creds._lookup_regional_access_boundary.side_effect = Exception(
            "Network failure"
        )

        worker = _regional_access_boundary_utils._RegionalAccessBoundaryRefreshThread(
            creds, request, rab_manager
        )
        worker.run()

        # Assert data was aggressively cleared rather than reused due to total expiration
        assert rab_manager._data.encoded_locations is None
        assert rab_manager._data.expiry is None
        assert rab_manager._data.cooldown_expiry is not None

    def test_regional_access_boundary_refresh_manager_start_refresh_safety_lock(self):
        manager = (
            _regional_access_boundary_utils._RegionalAccessBoundaryRefreshManager()
        )
        creds = mock.Mock()
        request = mock.Mock()
        rab_manager = mock.Mock()

        mock_worker = mock.Mock()
        mock_worker.is_alive.return_value = True
        manager._worker = mock_worker

        with mock.patch(
            "google.auth._regional_access_boundary_utils._RegionalAccessBoundaryRefreshThread"
        ) as mock_thread_class:
            manager.start_refresh(creds, request, rab_manager)

            mock_thread_class.assert_not_called()
            assert manager._worker == mock_worker


class AsyncCredentialsImpl(_credentials_async.CredentialsWithRegionalAccessBoundary):
    def __init__(self, universe_domain=None):
        super().__init__()
        if universe_domain:
            self._universe_domain = universe_domain

    async def _perform_refresh_token(self, request):
        self.token = "refreshed-token"
        self.expiry = (
            _helpers.utcnow()
            + _helpers.REFRESH_THRESHOLD
            + datetime.timedelta(seconds=5)
        )

    def with_quota_project(self, quota_project_id):
        raise NotImplementedError()

    def _build_regional_access_boundary_lookup_url(self, request=None):
        # Using self.token here to make the URL dynamic for testing purposes
        return "http://mock.url/lookup_for_{}".format(self.token)

    def _make_copy(self):
        new_credentials = self.__class__()
        self._copy_regional_access_boundary_manager(new_credentials)
        return new_credentials


class TestAsyncCredentialsWithRegionalAccessBoundary(object):
    @pytest.mark.asyncio
    async def test_maybe_start_refresh_async_blocking(self):
        creds = AsyncCredentialsImpl()
        creds._rab_manager._use_blocking_regional_access_boundary_lookup = True
        request = mock.Mock()

        with mock.patch.object(
            creds._rab_manager,
            "start_blocking_refresh_async",
            new_callable=mock.AsyncMock,
        ) as mock_start_blocking:
            await creds._maybe_start_regional_access_boundary_refresh_async(
                request, "http://example.com"
            )
            mock_start_blocking.assert_called_once_with(creds, request)

    @pytest.mark.asyncio
    async def test_start_blocking_refresh_async_success(self):
        creds = AsyncCredentialsImpl()
        request = mock.Mock()

        with mock.patch.object(
            creds,
            "_lookup_regional_access_boundary",
            new_callable=mock.AsyncMock,
            return_value={"encodedLocations": "0xABC"},
        ) as mock_lookup:
            await creds._rab_manager.start_blocking_refresh_async(creds, request)

            mock_lookup.assert_called_once_with(request, fail_fast=True)
            assert creds._rab_manager._data.encoded_locations == "0xABC"

    @pytest.mark.asyncio
    async def test_start_blocking_refresh_async_failure(self):
        creds = AsyncCredentialsImpl()
        request = mock.Mock()

        with mock.patch.object(
            creds,
            "_lookup_regional_access_boundary",
            new_callable=mock.AsyncMock,
            side_effect=Exception("error"),
        ) as mock_lookup:
            await creds._rab_manager.start_blocking_refresh_async(creds, request)

            mock_lookup.assert_called_once_with(request, fail_fast=True)
            assert creds._rab_manager._data.encoded_locations is None
            assert creds._rab_manager._data.cooldown_expiry is not None

    @pytest.mark.asyncio
    async def test_async_refresh_manager_session_closed_ignored(self):
        credentials = mock.AsyncMock()
        # Simulate a closed session RuntimeError when invoking the boundary lookup
        credentials._lookup_regional_access_boundary.side_effect = RuntimeError(
            "Session is closed"
        )

        request = mock.Mock()
        request._clone.return_value = request
        rab_manager = mock.Mock()

        manager = (
            _regional_access_boundary_utils._AsyncRegionalAccessBoundaryRefreshManager()
        )

        # Trigger refresh, which starts a background task that should swallow the error
        manager.start_refresh(credentials, request, rab_manager)

        # Wait for the background worker task to terminate
        await manager._worker_task

        # Verify that the lookup was still triggered but failed open cleanly
        credentials._lookup_regional_access_boundary.assert_called_once_with(request)
        rab_manager.process_regional_access_boundary_info.assert_called_once_with(None)

    @pytest.mark.asyncio
    async def test_start_refresh_async_clones_request_and_unwraps_partial(self):
        import functools

        credentials = mock.AsyncMock()
        credentials._lookup_regional_access_boundary.return_value = {
            "encodedLocations": "0xA30"
        }

        mock_request = mock.Mock()
        mock_cloned_request = mock.Mock()
        mock_request._clone.return_value = mock_cloned_request
        mock_cloned_request.close = mock.AsyncMock()

        # Wrap in a functools.partial to simulate AuthorizedSession.request() timeouts
        partial_request = functools.partial(mock_request, timeout=180)

        rab_manager = mock.Mock()

        manager = (
            _regional_access_boundary_utils._AsyncRegionalAccessBoundaryRefreshManager()
        )
        manager.start_refresh(credentials, partial_request, rab_manager)

        await manager._worker_task

        # Verify that actual_request._clone() was called
        mock_request._clone.assert_called_once()

        # Verify that the lookup ran on a re-wrapped partial of the cloned request
        called_arg = credentials._lookup_regional_access_boundary.call_args[0][0]
        assert isinstance(called_arg, functools.partial)
        assert called_arg.func is mock_cloned_request
        assert called_arg.keywords == {"timeout": 180}

        # Verify that the cloned request was closed cleanly in the finally block
        mock_cloned_request.close.assert_awaited_once()
        rab_manager.process_regional_access_boundary_info.assert_called_once_with(
            {"encodedLocations": "0xA30"}
        )

    @pytest.mark.asyncio
    async def test_start_refresh_suppresses_request_clone_exception(self):
        from google.auth import exceptions

        credentials = mock.AsyncMock()

        request = mock.Mock()
        request._clone.side_effect = exceptions.TransportError(
            "Cannot clone a closed transport."
        )

        rab_manager = mock.Mock()
        manager = (
            _regional_access_boundary_utils._AsyncRegionalAccessBoundaryRefreshManager()
        )

        manager.start_refresh(credentials, request, rab_manager)

        assert manager._worker_task is None
        credentials._lookup_regional_access_boundary.assert_not_called()
        rab_manager.process_regional_access_boundary_info.assert_called_once_with(None)

    @pytest.mark.asyncio
    async def test_start_refresh_async_mimics_ephemeral_session_closed_bug(self):
        # Specifically mimics the real-world race condition where a fast foreground main call
        # pulls the rug out from under the background worker when using an un-cloned session.
        import asyncio

        manager = (
            _regional_access_boundary_utils._AsyncRegionalAccessBoundaryRefreshManager()
        )

        worker_started_event = asyncio.Event()
        foreground_closed_event = asyncio.Event()

        class EphemeralRequest:
            def __init__(self):
                self.closed = False

            async def __call__(self, *args, **kwargs):
                worker_started_event.set()
                await foreground_closed_event.wait()
                if self.closed:
                    raise RuntimeError("Session is closed")
                return "success"

        ephemeral_req = EphemeralRequest()

        credentials = mock.AsyncMock()

        async def mock_lookup(req):
            return await req()

        credentials._lookup_regional_access_boundary.side_effect = mock_lookup

        rab_manager = mock.Mock()

        # Start the background refresh worker
        manager.start_refresh(credentials, ephemeral_req, rab_manager)

        # Wait until the background worker has actually started its speculative request
        await worker_started_event.wait()

        # Simulate fast foreground primary call closing the session
        ephemeral_req.closed = True
        foreground_closed_event.set()

        # Await the background worker task to settle
        await manager._worker_task

        # Verify that the background worker hit the "Session is closed" error and failed open cleanly
        rab_manager.process_regional_access_boundary_info.assert_called_once_with(None)


def test_get_service_account_rab_endpoint(monkeypatch):
    from google.auth.transport import _mtls_helper

    # Test Standard TLS
    monkeypatch.setattr(_mtls_helper, "check_use_client_cert", lambda: False)
    url = _regional_access_boundary_utils.get_service_account_rab_endpoint(
        "test@example.com"
    )
    assert (
        url
        == "https://iamcredentials.googleapis.com/v1/projects/-/serviceAccounts/test@example.com/allowedLocations"
    )

    # Test mTLS
    monkeypatch.setattr(_mtls_helper, "check_use_client_cert", lambda: True)
    url = _regional_access_boundary_utils.get_service_account_rab_endpoint(
        "test@example.com"
    )
    assert (
        url
        == "https://iamcredentials.mtls.googleapis.com/v1/projects/-/serviceAccounts/test@example.com/allowedLocations"
    )


def test_get_workforce_pool_rab_endpoint(monkeypatch):
    from google.auth.transport import _mtls_helper

    # Test Standard TLS
    monkeypatch.setattr(_mtls_helper, "check_use_client_cert", lambda: False)
    url = _regional_access_boundary_utils.get_workforce_pool_rab_endpoint("POOL_ID")
    assert (
        url
        == "https://iamcredentials.googleapis.com/v1/locations/global/workforcePools/POOL_ID/allowedLocations"
    )

    # Test mTLS
    monkeypatch.setattr(_mtls_helper, "check_use_client_cert", lambda: True)
    url = _regional_access_boundary_utils.get_workforce_pool_rab_endpoint("POOL_ID")
    assert (
        url
        == "https://iamcredentials.mtls.googleapis.com/v1/locations/global/workforcePools/POOL_ID/allowedLocations"
    )


def test_get_workload_identity_pool_rab_endpoint(monkeypatch):
    from google.auth.transport import _mtls_helper

    # Test Standard TLS
    monkeypatch.setattr(_mtls_helper, "check_use_client_cert", lambda: False)
    url = _regional_access_boundary_utils.get_workload_identity_pool_rab_endpoint(
        "PROJECT_NUM", "POOL_ID"
    )
    assert (
        url
        == "https://iamcredentials.googleapis.com/v1/projects/PROJECT_NUM/locations/global/workloadIdentityPools/POOL_ID/allowedLocations"
    )

    # Test mTLS
    monkeypatch.setattr(_mtls_helper, "check_use_client_cert", lambda: True)
    url = _regional_access_boundary_utils.get_workload_identity_pool_rab_endpoint(
        "PROJECT_NUM", "POOL_ID"
    )
    assert (
        url
        == "https://iamcredentials.mtls.googleapis.com/v1/projects/PROJECT_NUM/locations/global/workloadIdentityPools/POOL_ID/allowedLocations"
    )


def test_sync_refresh_manager_pickle():
    import pickle

    manager = _regional_access_boundary_utils._RegionalAccessBoundaryRefreshManager()
    manager._worker = mock.Mock()

    dumped = pickle.dumps(manager)
    loaded = pickle.loads(dumped)

    assert loaded._lock is not None
    assert loaded._worker is None


def test_manager_eq_different_type():
    manager = _regional_access_boundary_utils._RegionalAccessBoundaryManager()
    assert manager != "not a manager"


def test_set_initial_regional_access_boundary_empty():
    manager = _regional_access_boundary_utils._RegionalAccessBoundaryManager()
    manager.set_initial_regional_access_boundary(
        encoded_locations="", expiry=datetime.datetime.now()
    )
    assert manager._data.encoded_locations == ""
    assert manager._data.expiry is None


def test_set_initial_regional_access_boundary_with_value():
    manager = _regional_access_boundary_utils._RegionalAccessBoundaryManager()
    expiry = datetime.datetime.now()
    manager.set_initial_regional_access_boundary(
        encoded_locations="us-east1", expiry=expiry
    )
    assert manager._data.encoded_locations == "us-east1"
    assert manager._data.expiry == expiry


def test_sync_refresh_manager_start_refresh_executes():
    manager = _regional_access_boundary_utils._RegionalAccessBoundaryRefreshManager()
    creds = mock.Mock()
    request = mock.Mock()
    rab_manager = mock.Mock()

    with mock.patch(
        "google.auth._regional_access_boundary_utils._RegionalAccessBoundaryRefreshThread"
    ) as mock_thread_class:
        mock_thread = mock.Mock()
        mock_thread_class.return_value = mock_thread

        manager.start_refresh(creds, request, rab_manager)

        mock_thread_class.assert_called_once()
        mock_thread.start.assert_called_once()



# Copyright 2014 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import platform
from unittest import mock

import pytest

from google.auth import metrics
from google.auth import version


def test_add_metric_header():
    headers = {}
    metrics.add_metric_header(headers, None)
    assert headers == {}

    headers = {"x-goog-api-client": "foo"}
    metrics.add_metric_header(headers, "bar")
    assert headers == {"x-goog-api-client": "foo bar"}

    headers = {}
    metrics.add_metric_header(headers, "bar")
    assert headers == {"x-goog-api-client": "bar"}


@mock.patch.object(platform, "python_version", return_value="<python-version>")
def test_versions(mock_python_version):
    version_save = version.__version__
    version.__version__ = "<library-version>"
    assert (
        metrics.python_and_auth_lib_version()
        == "gl-python/<python-version> auth/<library-version>"
    )
    version.__version__ = version_save


@pytest.mark.parametrize(
    "func, expected_suffix",
    [
        (metrics.token_request_access_token_mds, "auth-request-type/at cred-type/mds"),
        (metrics.token_request_id_token_mds, "auth-request-type/it cred-type/mds"),
        (
            metrics.token_request_access_token_impersonate,
            "auth-request-type/at cred-type/imp",
        ),
        (
            metrics.token_request_id_token_impersonate,
            "auth-request-type/it cred-type/imp",
        ),
        (
            metrics.token_request_access_token_sa_assertion,
            "auth-request-type/at cred-type/sa",
        ),
        (
            metrics.token_request_id_token_sa_assertion,
            "auth-request-type/it cred-type/sa",
        ),
        (metrics.token_request_user, "cred-type/u"),
        (metrics.mds_ping, "auth-request-type/mds"),
        (metrics.reauth_start, "auth-request-type/re-start"),
        (metrics.reauth_continue, "auth-request-type/re-cont"),
    ],
)
@mock.patch(
    "google.auth.metrics.python_and_auth_lib_version",
    return_value="gl-python/<python-version> auth/<library-version>",
)
def test_metric_values(mock_python_and_auth_lib_version, func, expected_suffix):
    # mock_python_and_auth_lib_version is injected by mock.patch but is not
    # explicitly referenced in the test body as the mock behaves as configured.
    expected = (
        f"gl-python/<python-version> auth/<library-version> {expected_suffix}".strip()
    )
    assert func() == expected


@mock.patch(
    "google.auth.metrics.python_and_auth_lib_version",
    return_value="gl-python/<python-version> auth/<library-version>",
)
def test_byoid_metric_header(mock_python_and_auth_lib_version):
    # mock_python_and_auth_lib_version is injected by mock.patch but is not
    # explicitly referenced in the test body as the mock behaves as configured.
    metrics_options = {}
    assert (
        metrics.byoid_metrics_header(metrics_options)
        == "gl-python/<python-version> auth/<library-version> google-byoid-sdk"
    )
    metrics_options["testKey"] = "testValue"
    assert (
        metrics.byoid_metrics_header(metrics_options)
        == "gl-python/<python-version> auth/<library-version> google-byoid-sdk testKey/testValue"
    )



# Copyright 2014 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os

from setuptools import find_namespace_packages
from setuptools import setup

cryptography_base_require = [
    "cryptography >= 38.0.3; python_version < '3.14'",
    "cryptography >= 41.0.5; python_version >= '3.14'",
]

DEPENDENCIES = (
    "pyasn1-modules>=0.2.1",
    *cryptography_base_require,
)

requests_extra_require = ["requests >= 2.30.0, < 3.0.0"]

aiohttp_extra_require = [
    "aiohttp >= 3.8.0, < 4.0.0; python_version < '3.14'",
    "aiohttp >= 3.9.0, < 4.0.0; python_version >= '3.14'",
    *requests_extra_require,
]

pyjwt_extra_require = ["pyjwt>=2.0"]

reauth_extra_require = ["pyu2f>=0.1.5"]

enterprise_cert_extra_require = cryptography_base_require

urllib3_extra_require = [
    "urllib3 >= 1.26.15, < 3.0.0",
    "packaging >= 20.0",
]

rsa_extra_require = ["rsa>=4.0.0,<5"]

grpc_extra_require = [
    "grpcio >= 1.59.0, < 2.0.0; python_version < '3.14'",
    "grpcio >= 1.75.1, < 2.0.0; python_version >= '3.14'",
]

# Unit test requirements.
testing_extra_require = [
    *grpc_extra_require,
    "flask",
    "freezegun",
    *pyjwt_extra_require,
    "pytest",
    "pytest-cov",
    "pytest-localserver",
    *reauth_extra_require,
    "responses",
    *urllib3_extra_require,
    # Async Dependencies
    *aiohttp_extra_require,
    "aioresponses",
    "pytest-asyncio",
]

extras = {
    # Note: cryptography was made into a required dependency. Extra is kept for backwards compatibility
    "cryptography": cryptography_base_require,
    # pyopenssl is deprecated, kept for backwards compatibility
    "pyopenssl": cryptography_base_require,
    "aiohttp": aiohttp_extra_require,
    "enterprise_cert": enterprise_cert_extra_require,
    "pyjwt": pyjwt_extra_require,
    "reauth": reauth_extra_require,
    "requests": requests_extra_require,
    "testing": testing_extra_require,
    "urllib3": urllib3_extra_require,
    "rsa": rsa_extra_require,
    "grpc": grpc_extra_require,
}

with io.open("README.rst", "r") as fh:
    long_description = fh.read()

package_root = os.path.abspath(os.path.dirname(__file__))

version = {}
with open(os.path.join(package_root, "google/auth/version.py")) as fp:
    exec(fp.read(), version)
version = version["__version__"]

setup(
    name="google-auth",
    version=version,
    author="Google Cloud Platform",
    author_email="googleapis-packages@google.com",
    description="Google Authentication Library",
    long_description=long_description,
    url="https://github.com/googleapis/google-cloud-python/tree/main/packages/google-auth",
    packages=find_namespace_packages(
        exclude=("tests*", "system_tests*", "docs*", "samples*")
    ),
    package_data={"google.auth": ["py.typed"], "google.oauth2": ["py.typed"]},
    install_requires=DEPENDENCIES,
    extras_require=extras,
    python_requires=">=3.10",
    license="Apache 2.0",
    keywords="google auth oauth client",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP",
    ],
)
